//
// SierpinskiSolidsView.cpp : implementation of the CSierpinskiSolidsView class
//

#include "stdafx.h"
#include "SierpinskiSolids.h"

#include "SierpinskiSolidsDoc.h"
#include "SierpinskiSolidsView.h"
#include <math.h>

#define M_PI 3.141526

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// add support for OpenGL 1.1 if we're using an old header
// These are new PIXELFORMATDESCRIPTOR flags for OpenGL 1.1

#ifndef PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED		0x00001000
#endif

#ifndef PFD_DEPTH_DONTCARE
#define PFD_DEPTH_DONTCARE			0x20000000
#endif

static const float m_ROTSCALE = 90.0;
static const float m_ZOOMSCALE = 0.008;
static GLfloat objectXform[4][4] = {
	{1.0, 0.0, 0.0, 0.0},
	{0.0, 1.0, 0.0, 0.0},
	{0.0, 0.0, 1.0, 0.0},
	{0.0, 0.0, 0.0, 1.0}
};



/////////////////////////////////////////////////////////////////////////////
// CSierpinskiSolidsView

IMPLEMENT_DYNCREATE(CSierpinskiSolidsView, CView)

BEGIN_MESSAGE_MAP(CSierpinskiSolidsView, CView)
	//{{AFX_MSG_MAP(CSierpinskiSolidsView)
	ON_WM_SIZE()
	ON_WM_CANCELMODE()
	ON_WM_ERASEBKGND()
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_WM_CHAR()
	ON_COMMAND(ID_Sponge, OnSponge)
	ON_COMMAND(ID_Tetra, OnTetra)
	ON_COMMAND(ID_RESET_VIEW, OnResetView)
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSierpinskiSolidsView construction/destruction

CSierpinskiSolidsView::CSierpinskiSolidsView()
{
	// TODO: add construction code here
	MAX_Level = 7;
	Movement = NONE;
	Shadows = TRUE;
	Reflections = FALSE;
	Spin = FALSE;
	rotAngle = 0.0;
	m_RotStep = 5.0;
}

CSierpinskiSolidsView::~CSierpinskiSolidsView()
{
}


/////////////////////////////////////////////////////////////////////////////
// CSierpinskiSolidsView printing

BOOL CSierpinskiSolidsView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CSierpinskiSolidsView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CSierpinskiSolidsView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CSierpinskiSolidsView diagnostics

#ifdef _DEBUG
void CSierpinskiSolidsView::AssertValid() const
{
	CView::AssertValid();
}

void CSierpinskiSolidsView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CSierpinskiSolidsDoc* CSierpinskiSolidsView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CSierpinskiSolidsDoc)));
	return (CSierpinskiSolidsDoc*)m_pDocument;
}
#endif //_DEBUG




BOOL CSierpinskiSolidsView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	cs.style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CSierpinskiSolidsView drawing

void CSierpinskiSolidsView::OnDraw(CDC* pDC)
{
	CSierpinskiSolidsDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// Make the HGLRC current.
	BOOL bResult = wglMakeCurrent (pDC->m_hDC, m_hrc);
	if (!bResult)
	{
	  TRACE("wglMakeCurrent Failed %x\r\n", GetLastError() ) ;
	}

	//
	// Draw the scene
	//
	glMatrixMode( GL_MODELVIEW );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	MAX_Level = pDoc->Recursion_Level;

	if( Shadows )
	{
		//
		//  Initialize the random number generator to get consistent results
		//    for each frame or each pass thru with multi-pass techniques.
		//
		srand( 0 );
		//
		//  Draw the scene flattened onto the ground plane.
		//
		DrawShadows( pDoc );
	}

	//
	//  Initialize the random number generator to get consistent results
	//    for each frame or each pass thru with multi-pass techniques.
	//
	srand( 0 );

	if( pDoc->Draw_Tetra )
		SierpinskiGen( pDoc->BaseTetra, 0 );


	if( pDoc->Draw_Sponge )
		SierpinskiGenCube( pDoc->BaseCube, 0 );
	
	glFinish();

	// Swap buffers.
	SwapBuffers(pDC->m_hDC) ;

	//
	// If we are in a self-animated mode, update the animation parameters
	//    and force a redraw.
	//
	if( Spin )
	{
		rotAngle += 2.0;
		Invalidate( TRUE );
	}

}

void CSierpinskiSolidsView::DrawShadows( CSierpinskiSolidsDoc* pDoc )
{
	//
	// Set-up the transformation matrix to flatten the scene onto the ground plane.
	////////////  Note, this does not work yet. ////////////////
	//
	
	glFrontFace( GL_CW );
	glPushMatrix();
	glScalef( 1, 1, -0.99 );
	glTranslatef( 0, 0, 2.4 );

	if( pDoc->Draw_Tetra )
		SierpinskiGen( pDoc->BaseTetra, 0 );


	if( pDoc->Draw_Sponge )
		SierpinskiGenCube( pDoc->BaseCube, 0 );

	glPopMatrix();
	glFrontFace( GL_CCW );
	//
	// Draw the ground plane
	//
	glDisable( GL_LIGHTING );
	glDisable( GL_DEPTH_TEST );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glColor4f( 0.7, 0.7, 0.7, 0.8 );
	glBegin( GL_QUADS );
		glVertex3f( -4.0, -4.0, -1.2 );
		glVertex3f(  4.0, -4.0, -1.2 );
		glVertex3f(  4.0,  4.0, -1.2 );
		glVertex3f( -4.0,  4.0, -1.2 );
	glEnd();
	glEnable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );
	glDisable( GL_BLEND );


}

void CSierpinskiSolidsView::SierpinskiGen(Tetrahedron &tetra, int level)
{
	//
	// If we reached the maximum recursion level,
	//    stop and draw the tetrahedron.
	// If animation is turned on, the rotAngle will force the
	//   tetrahedron to spin.
	//
	if( level > MAX_Level )
		tetra.DrawFaces( rotAngle*level );
	//
	// Otherwise, split the tetrahedron into 5 smaller tetrahedra,
	//    Drawing the 4 that are on the corners.
	// Logically, we split each edge into two.
	// Build the tetra connecting each original vertex with the
	// midpoints of the edges it shares.
	//
	else
	{
		Tetrahedron subtetra;
		//
		// Tetrahedron at the first vertex.
		//
		subtetra.Pts[0] = tetra.Pts[0];
		subtetra.Pts[1] = tetra.mid[0];
		subtetra.Pts[2] = tetra.mid[4];
		subtetra.Pts[3] = tetra.mid[3];
		//
		// The Tetrahedron class contains information on the 
		//    edge midpoints. Bad programming here, in that this
		//    class implicitly knows the ordering of the edges.
		//   ------- Fix later --------
		//         Add a method to Tetrahedron that subdivides itself
		//           and passes back the resulting five tetrahedra.
		//
		subtetra.CalcMidpts();
		//
		// Recursively subdivide each tetra.
		// Increment the level by one to note we are
		//   one level further in the recursion that our
		//   parent.
		// Some permutation (acceleration) on the level is
		//   added to force some nodes to terminate earlier
		//   than others.
		//
		SierpinskiGen( subtetra, (level+1) );

		glColor3f( 1.0, 1.0, 0.0 );
		//
		// Tetrahedron at the second vertex.
		//
		subtetra.Pts[0] = tetra.Pts[1];
		subtetra.Pts[1] = tetra.mid[1];
		subtetra.Pts[2] = tetra.mid[0];
		subtetra.Pts[3] = tetra.mid[5];
		subtetra.CalcMidpts();
		SierpinskiGen( subtetra, (level+2) );

		glColor3f( 0.0, 1.0, 1.0 );
		//
		// Tetrahedron at the third vertex.
		//
		subtetra.Pts[0] = tetra.Pts[2];
		subtetra.Pts[1] = tetra.mid[4];
		subtetra.Pts[2] = tetra.mid[1];
		subtetra.Pts[3] = tetra.mid[2];
		subtetra.CalcMidpts();
		SierpinskiGen( subtetra, (level+3) );

		glColor3f( 1.0, 0.0, 1.0 );
		//
		// Tetrahedron at the fourth vertex.
		//
		subtetra.Pts[3] = tetra.Pts[3];
		subtetra.Pts[0] = tetra.mid[3];
		subtetra.Pts[1] = tetra.mid[5];
		subtetra.Pts[2] = tetra.mid[2];
		subtetra.CalcMidpts();
		SierpinskiGen( subtetra, (level+1) );

	}
}

void CSierpinskiSolidsView::SierpinskiGenCube(Cube &cube1, const int level)
{
	//
	// If we reached the maximum recursion level,
	//    stop and draw the cube.
	//
	if( level > MAX_Level )
		cube1.DrawFaces();
	//
	// Otherwise, split the cube into 27 smaller cubes,
	//    Drawing the 20 that are on the edges.
	// Logically, this give us a 4x4x4 grid of points.
	// Build the cubes for all of these, except for:
	//   1. The center interior cube.
	//   2. The six cubes which contain a face in the center of the 
	//        original cubes face.
	//
	else
	{
		//
		// The Cube class contains information on how to subdivide itself
		//    into a 4x4x4 set of vertices.
		//
		cube1.CalcGrid();
		Cube subCube;
		for( int i = 0; i < 3; i ++ )
		{
			for( int j = 0; j < 3; j++ )
			{
				for( int k = 0; k < 3; k++ )
				{
					if( ((i==1)&&(j==1)) || ((j==1)&&(k==1)) || ((i==1)&&(k==1)) )
					{
						//
						// If any two dimensions indicate the start of the center 
						//   division point, then we have an interior cube and
						//   do nothing.
						//
						;
					}
					else
					{
						//
						// Otherwise, we build the cube from the 4x4x4 grid
						//
						subCube.Pts[0] = cube1.Grid[i][j][k];
						subCube.Pts[1] = cube1.Grid[i+1][j][k];
						subCube.Pts[2] = cube1.Grid[i+1][j+1][k];
						subCube.Pts[3] = cube1.Grid[i][j+1][k];
						subCube.Pts[4] = cube1.Grid[i][j][k+1];
						subCube.Pts[5] = cube1.Grid[i+1][j][k+1];
						subCube.Pts[6] = cube1.Grid[i+1][j+1][k+1];
						subCube.Pts[7] = cube1.Grid[i][j+1][k+1];
						//
						// Recursively subdivide each subcube.
						// Increment the level by one to note we are
						//   one level further in the recursion that our
						//   parent.
						// Some permutation (acceleration) on the level is
						//   added to force some nodes to terminate earlier
						//   than others.
						//
						SierpinskiGenCube( subCube, (level+1 + ((i+j+k)%2) ) );
					}
				}
			}
		}
	}

}




/////////////////////////////////////////////////////////////////////////////
// CSierpinskiSolidsView message handlers

void CSierpinskiSolidsView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
	// TODO: Add your message handler code here
	CClientDC dc(this) ;

   BOOL bResult = wglMakeCurrent (dc.m_hDC, m_hrc);
   if (!bResult)
   {
      TRACE("wglMakeCurrent Failed %x\r\n", GetLastError() ) ;
      return ;
   }

   //
   // Determine the new aspect ratio
   //
   GLdouble gldAspect = (GLdouble) cx/ (GLdouble) cy;
   //
   // Reset the projection matrix with the new aspect ratio.
   //
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(40.0, gldAspect, 1.001, 60.0);
   glTranslatef( 0.0, 0.0, -25.0 );
   //
   // Set the viewport to take up the entire window.
   //
   glViewport(0, 0, cx, cy);
   windowSize.x = cx;
   windowSize.y = cy;

   //wglMakeCurrent(NULL, NULL);
}

void CSierpinskiSolidsView::OnCancelMode() 
{
	CView::OnCancelMode();
	
	// TODO: Add your message handler code here
	
}

void CSierpinskiSolidsView::OnDestroy() 
{
	CView::OnDestroy();
	
	// TODO: Add your message handler code here
	wglMakeCurrent(NULL, NULL) ;
	if (m_hrc)
	{
	   wglDeleteContext(m_hrc) ;
	   m_hrc = NULL ;
	}	
	
}

BOOL CSierpinskiSolidsView::OnEraseBkgnd(CDC* pDC) 
{
	// TODO: Add your message handler code here and/or call default
	//return CView::OnEraseBkgnd( pDC );
	return TRUE;
}

int CSierpinskiSolidsView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// TODO: Add your specialized creation code here
   CClientDC dc(this) ;
   //
   // Fill in the pixel format descriptor.
   //
   static PIXELFORMATDESCRIPTOR pfd ;
	   memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR)) ;  // Initialize all entries to zero.
	   pfd.nSize      = sizeof(PIXELFORMATDESCRIPTOR); 
	   pfd.nVersion   = 1 ; 
	   pfd.dwFlags    = PFD_DOUBLEBUFFER |      // Allocate space for double buffering.
						PFD_SUPPORT_OPENGL |    // We will be using OpenGL
						PFD_DRAW_TO_WINDOW ;    // We want it to be visible, draw to the graphics device.
	   pfd.iPixelType = PFD_TYPE_RGBA ;         // Allocate an RGBA display
	   pfd.cColorBits = 24 ;
	   pfd.cDepthBits = 32 ;
	   pfd.iLayerType = PFD_MAIN_PLANE ;
	   pfd.cStencilBits = 1;                    // We also want a stencil buffer. We only need one bit.

   int nPixelFormat = ChoosePixelFormat(dc.GetSafeHdc(), &pfd);
   if (nPixelFormat == 0)
   {
      TRACE("ChoosePixelFormat Failed %d\r\n",GetLastError()) ;
      return -1 ;
   }
   TRACE("Pixel Format %d\r\n",nPixelFormat) ;

   BOOL bResult = SetPixelFormat (dc.GetSafeHdc(), nPixelFormat, &pfd);
   if (!bResult)
   {
      TRACE("SetPixelFormat Failed %d\r\n",GetLastError()) ;
      return -1 ;
   }

   //
   // Create a rendering context.
   //
   m_hrc = wglCreateContext(dc.GetSafeHdc());
   if (!m_hrc)
   {
      TRACE("wglCreateContext Failed %x\r\n", GetLastError()) ;
      return -1;
   }
	wglMakeCurrent( dc.GetSafeHdc(), m_hrc );
   //
   // Check to make sure we do not need a color pallette
   //
   /*
   if( pfd.dwFlags & PFD_NEED_PALETTE )
   {
	   TRACE( "Only 24-bit color mode is supported\n" );
	   return -1;
   }
   */

	//
	// Set up OpenGL state
	//
	// Initialize the viewing matrix.
	//
   glMatrixMode( GL_MODELVIEW );
   glLoadIdentity();

   //
   // Initialize the lights and lighting state.
   //
   const float light2_pos[3] = {100.0, 0.0, 100.0 }; 
   glEnable( GL_LIGHTING);
   glEnable( GL_LIGHT0 );
   glEnable( GL_LIGHT1 );
   glEnable( GL_LIGHT2 );
   glLightfv( GL_LIGHT1, GL_POSITION, light2_pos );
   float ambient[3] = {0.1, 0.1, 0.1};
   glLightfv( GL_LIGHT2, GL_AMBIENT, ambient );

   //
   // Turn on the z-buffer
   //
   glEnable( GL_DEPTH_TEST );
   //
   // Since we were diligent in correctly defining al clockwise faces, 
   //   enable the backface culling for better efficiency.
   //
   glEnable( GL_CULL_FACE );

   //
   // Set the background color
   //
   const float red = 0.6;
   glClearColor( red, 0.0, 0.0, 1.0 );

	glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    glEnable( GL_COLOR_MATERIAL );
	glShadeModel( GL_FLAT );
 
	return 0;
}


///////////////////////////////////////
//  User events
//
void CSierpinskiSolidsView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// 
	// Turn on user interactive rotations.
	// As the user moves the mouse, the scene will rotate.
	//
	Movement = ROTATE;

	//
	// Map the mouse position to a logical sphere location.
	// Keep it in the class variable lastPoint.
	//
	lastPoint = trackBallMapping( point );
	//
	// Make sure we are modifying the MODELVIEW matrix.
	//
	glMatrixMode( GL_MODELVIEW );
	
	CView::OnLButtonDown(nFlags, point);
}

void CSierpinskiSolidsView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	// 
	// Turn-off the rotations.
	//
	Movement = NONE;
	
	CView::OnLButtonUp(nFlags, point);
}

void CSierpinskiSolidsView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// 
	// Turn on user interactive zooming.
	// As the user moves the mouse, the scene will zoom in or out
	//   depending on the x-direction of travel.
	//
	Movement = ZOOM;
	//
	// Set the last point, so future mouse movements can determine
	//   the distance moved.
	//
	lastPoint.x = (float) point.x;
	lastPoint.y = (float) point.y;
	//
	// Make sure we are modifying the PROJECTION matrix.
	//
	glMatrixMode( GL_PROJECTION );
	
	CView::OnRButtonDown(nFlags, point);
}

void CSierpinskiSolidsView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default	
	Movement = NONE;

	CView::OnRButtonUp(nFlags, point);
}

void CSierpinskiSolidsView::OnMouseMove(UINT nFlags, CPoint point) 
{
	// 
	// Handle any necessary mouse movements
	//
	Vec3f direction;
	float pixel_diff;
	float rot_angle, zoom_factor;
	Vec3f curPoint;

	switch (Movement) 
	{
		case ROTATE :  // Left-mouse button is being held down
			{
				curPoint = trackBallMapping( point );  // Map the mouse position to a logical sphere location.
				direction = curPoint - lastPoint;
				float velocity = direction.Length();
				if( velocity > 0.0001 )
				{
					//
					// Rotate about the axis that is perpendicular to the great circle connecting the mouse movements.
					//
					Vec3f rotAxis;
					rotAxis.crossProd( lastPoint, curPoint );
					rot_angle = velocity * m_ROTSCALE;
					//
					// We need to apply the rotation as the last transformation.
					//   1. Get the current matrix and save it.
					//   2. Set the matrix to the identity matrix (clear it).
					//   3. Apply the trackball rotation.
					//   4. Pre-multiply it by the saved matrix.
					//
					glGetFloatv( GL_MODELVIEW_MATRIX, (GLfloat *) objectXform );
					glLoadIdentity();
					glRotatef( rot_angle, rotAxis.x, rotAxis.y, rotAxis.z );
					glMultMatrixf( (GLfloat *) objectXform );
					//
					//  If we want to see it, we need to force the system to redraw the scene.
					//
					Invalidate( TRUE );
				}
				break;
			}
		case ZOOM :  // Right-mouse button is being held down
			//
			// Zoom into or away from the scene based upon how far the mouse moved in the x-direction.
			//   This implementation does this by scaling the eye-space.
			//   This should be the first operation performed by the GL_PROJECTION matrix.
			//   1. Calculate the signed distance
			//       a. movement to the left is negative (zoom out).
			//       b. movement to the right is positive (zoom in).
			//   2. Calculate a scale factor for the scene s = 1 + a*dx
			//   3. Call glScalef to have the scale be the first transformation.
			// 
			pixel_diff = point.x - lastPoint.x; 
			zoom_factor = 1.0 + pixel_diff * m_ZOOMSCALE;
			glScalef( zoom_factor, zoom_factor, zoom_factor );
			//
			// Set the current point, so the lastPoint will be saved properly below.
			//
			curPoint.x = (float) point.x;  curPoint.y = (float) point.y;  (float) curPoint.z = 0;
			//
			//  If we want to see it, we need to force the system to redraw the scene.
			//
			Invalidate( TRUE );
			break;
	}
	//
	// Save the location of the current point for the next movement. 
	//
	lastPoint = curPoint;
	
	CView::OnMouseMove(nFlags, point);
}


void CSierpinskiSolidsView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// 
	// Handle the following keyboard events:
	//    +  => increase the recursion level by one.
	//    =  =>   same as +, to make it case insensitive.
	//    -  => decrease the recursion level by one.
	//    _  =>   same as -, to make it case insensitive.
	//    s  => toggle the shadow plane on and off.
	//    r  => toggle the reflection plane on and off.
	//    a  => animate the scene by spinning each tetrahedron in place.
	//
	//    8  => up-arrow on numeric keypad. Rotate counter-clockwise about the x-axis in eye-space
	//    k  =>   same as 8 (UNIX-vi movement)
	//    2  => down-arrow on numeric keypad. Rotate clockwise about the x-axis in eye-space
	//    j  =>   same as 2
	//    4  => left-arrow on numeric keypad. Rotate counter-clockwise about the y-axis in eye-space
	//    h  =>   same as 4
	//    6  => right-arrow on numeric keypad. Rotate clockwise about the y-axis in eye-space
	//    l  =>   same as 6
	//
	if( (nChar =='+') || (nChar == '=') )
	{
		CSierpinskiSolidsDoc* pDoc = GetDocument();
		ASSERT_VALID(pDoc);
		if( pDoc->Recursion_Level < pDoc->Max_Recursion_Level )
			pDoc->Recursion_Level++;
		Invalidate( TRUE );
	}
	if( (nChar =='-') || (nChar == '_') )
	{
		CSierpinskiSolidsDoc* pDoc = GetDocument();
		ASSERT_VALID(pDoc);
		if( pDoc->Recursion_Level > 0 )
			pDoc->Recursion_Level--;
		Invalidate( TRUE );
	}
	if( nChar == 's' )
	{
		Shadows ^= TRUE;
		Invalidate();
	}
	if( nChar == 'r' )
	{
		Reflections ^= TRUE;
		Invalidate();
	}
	if( nChar == 'a' )
	{
		Spin ^= TRUE;
		Invalidate();
	}
	if( (nChar == '8') || (nChar == 'k') )
	{
		Rotate_x( -m_RotStep );
		Invalidate( TRUE );
	}
	if( nChar == '2' || (nChar == 'j') )
	{
		Rotate_x( m_RotStep );
		Invalidate( TRUE );
	}
	if( nChar == '4' || (nChar == 'h') )
	{
		Rotate_y( -m_RotStep );
		Invalidate( TRUE );
	}
	if( nChar == '6' || (nChar == 'l') )
	{
		Rotate_y( m_RotStep );
		Invalidate( TRUE );
	}
	

	CView::OnChar(nChar, nRepCnt, nFlags);
}

//
// Utility routine to calculate the 3D position of a 
// projected unit vector onto the xy-plane. Given any
// point on the xy-plane, we can think of it as the projection
// from a sphere down onto the plane. The inverse is what we
// are after.
//
Vec3f CSierpinskiSolidsView::trackBallMapping(CPoint point)
{
	Vec3f v;
	float d;

	v.x = (2.0*point.x - windowSize.x) / windowSize.x;
	v.y = (windowSize.y - 2.0*point.y) / windowSize.y;
	v.z = 0.0;
	d = v.Length();
	d = (d < 1.0) ? d : 1.0;  // If d is > 1, then clamp it at one.
	v.z = sqrtf( 1.001 - d*d );  // project the line segment up to the surface of the sphere.

	v.Normalize();  // We forced d to be less than one, not v, so need to normalize somewhere.


	return v;
}


void CSierpinskiSolidsView::OnSponge() 
{
	// 
	// Switch over to drawing a Monger Sponge.
	//
	CSierpinskiSolidsDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	pDoc->Draw_Sponge = TRUE;
	pDoc->Draw_Tetra = FALSE;
	//
	//  If we want to see it, we need to force the system to redraw the scene.
	//
	Invalidate( TRUE );
	
}

void CSierpinskiSolidsView::OnTetra() 
{
	// 
	// Switch over to drawing a Sierpinski Tetrahedron.
	//
	CSierpinskiSolidsDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	pDoc->Draw_Sponge = FALSE;
	pDoc->Draw_Tetra = TRUE;
	//
	//  If we want to see it, we need to force the system to redraw the scene.
	//
	Invalidate( TRUE );
	
}

void CSierpinskiSolidsView::OnResetView() 
{
	// 
	// Reset the view to the original
	//
	glLoadIdentity();
	//
	//  If we want to see it, we need to force the system to redraw the scene.
	//
	Invalidate( TRUE );
	
}

void CSierpinskiSolidsView::Rotate_x(const float angle)
{
	//
	// We need to apply the rotation as the last transformation.
	//   1. Get the current matrix and save it.
	//   2. Set the matrix to the identity matrix (clear it).
	//   3. Apply the trackball rotation.
	//   4. Pre-multiply it by the saved matrix.
	//
	glGetFloatv( GL_MODELVIEW_MATRIX, (GLfloat *) objectXform );
	glLoadIdentity();
	glRotatef( angle, 1.0, 0.0, 0.0 );
	glMultMatrixf( (GLfloat *) objectXform );
}

void CSierpinskiSolidsView::Rotate_y(const float angle)
{
	//
	// We need to apply the rotation as the last transformation.
	//   1. Get the current matrix and save it.
	//   2. Set the matrix to the identity matrix (clear it).
	//   3. Apply the trackball rotation.
	//   4. Pre-multiply it by the saved matrix.
	//
	glGetFloatv( GL_MODELVIEW_MATRIX, (GLfloat *) objectXform );
	glLoadIdentity();
	glRotatef( angle, 0.0, 1.0, 0.0 );
	glMultMatrixf( (GLfloat *) objectXform );
}
